@@ -10,12 +10,13 @@
 #  Copyright  1999, Kazuyuki Okamoto (euc-jp support in sgml2txt, sgml2html, and sgml2latex)
 #  Copyright  1999, Tetsu ONO (euc-jp support in sgml2txt, sgml2html, and sgml2latex)
 #  Copyright  2000, Juan Jose Amor (Support for PDF files)
-#  Copyright  2006-2009, Agustin Martin
+#  Copyright  2006-2020, Agustin Martin
 # ------------------------------------------------------------------
 
 package LinuxDocTools::fmt_latex2e;
 use strict;
 
+use LinuxDocTools::Utils qw(ldt_log);
 use LinuxDocTools::CharEnts;
 use LinuxDocTools::Vars;
 use LinuxDocTools::Lang;
@@ -52,7 +53,10 @@ $latex2e->{OPTIONS} = [
 		       { option => "latex",
 			 type => "l",
 			 'values' => [ "latex", "hlatexp", "platex", "jlatex" ],
-			 short => "x" }
+			 short => "x" },
+		       { option => "verbosity",
+			 type => "i",
+			 short => "V" }
 		       ];
 $latex2e->{output}         = "tex";
 $latex2e->{pagenumber}     = 1;
@@ -61,32 +65,31 @@ $latex2e->{bibtex}         = 0;
 $latex2e->{makeindex}      = 0;
 $latex2e->{latex}          = "unknown";
 $latex2e->{dvips}          = "unknown";
+$latex2e->{verbosity}      = -1;
 $Formats{$latex2e->{NAME}} = $latex2e;
 
+sub ldt_cat {
+  my $file    = shift;
+
+  open ( my $INPUT, "<$file")
+    or print STDERR "ERROR: Could not open \"$file\" for read\n";
+  while (<$INPUT>){
+    print STDERR $_;
+  }
+  close $INPUT;
+}
+
 # ------------------------------------------------------------------
 $latex2e->{preNSGMLS} = sub {
-# ------------------------------------------------------------------
+  # ----------------------------------------------------------------
   $global->{NsgmlsOpts} .= " -ifmttex ";
+  my $msgheader = "fmt_latex2e::preNSGMLS";
 
-  #  for Japanese jlatex users
-  if ($global->{language} eq "ja" && $latex2e->{latex} eq "unknown") {
-    $latex2e->{latex} = "jlatex";
-    $latex2e->{dvips} = "dvi2ps";
-    # for Japanese platex users
-    #	$latex2e->{latex} = "platex";
-    #	$latex2e->{dvips} = "dvips";
-  }
+  # NsgmlsPrePipe command
+  my $NsgmlsPrePipe_command = "cat $global->{file} | sed 's/_/\\&lowbar;/g' ";
+  ldt_log "  ${msgheader}::NsgmlsPrePipe_command\n    $NsgmlsPrePipe_command";
 
-  # for Korean users
-  if ($global->{language} eq "ko" && $latex2e->{latex} eq "unknown") {
-    $latex2e->{latex} = "hlatexp";
-  }
-
-  # default process command
-  $latex2e->{latex} = "latex" if ($latex2e->{latex} eq "unknown");
-  $latex2e->{dvips} = "dvips" if ($latex2e->{dvips} eq "unknown");
-
-  $global->{NsgmlsPrePipe} = "cat $global->{file} | sed 's/_/\\&lowbar;/g' ";
+  $global->{NsgmlsPrePipe} = $NsgmlsPrePipe_command;
 };
 
 # extra `\\' here for standard `nsgmls' output
@@ -108,8 +111,11 @@ $latex2e_escapes{'|'} = '{$|$}';
 my $in_verb;
 my $remove_comment; # added 2000 Jan 25 by t.sano
 
-# passed to `parse_data' below in latex2e_preASP
+# ------------------------------------------------------------------
 my $latex2e_escape = sub {
+  # ----------------------------------------------------------------
+  # passed to `parse_data' below in latex2e_preASP
+  # ----------------------------------------------------------------
   my ($data) = @_;
 
   if (!$in_verb) {
@@ -218,9 +224,133 @@ sub latex2e_defnam($) {
     chr ($anum + ($num / 26 % 26)) .
     chr ($anum + ($num % 26));
 
+  ldt_log "  fmt_latex2e::latex2e_defnam:\    num: $num, defnam: $defnam";
   return ($defnam);
 };
 
+# ------------------------------------------------------------------
+sub latex2e_fine_tune_options {
+  # ----------------------------------------------------------------
+  # Set defaults and fine tune $latex2e->{latex,dvips,output}
+  # ----------------------------------------------------------------
+  my $latex2e = shift;
+  my $msgheader = "fmt_latex2e::latex2e_fine_tune_options";
+
+  ldt_log "  ${msgheader} general info before fine tuning:";
+  ldt_log "    format: $global->{format}, charset: $global->{charset}, lang: $global->{language}";
+  ldt_log "    latex: $latex2e->{latex}, dvips: $latex2e->{dvips}, output: $latex2e->{output}, verbosity: $latex2e->{verbosity}";
+
+  # Set LaTeX verbosity if not explicitly set
+  if ( $latex2e->{verbosity} == -1 ){
+    $latex2e->{verbosity} = ( $global->{debug} ) ? 1 : 0;
+  }
+
+  # Fine tune $latex2e->{latex} for defaults
+  if ( $latex2e->{latex} eq "unknown" ) {
+    if ($global->{language} eq "ja"){
+      #  for Japanese jlatex users
+      $latex2e->{latex} = "jlatex";
+      # for Japanese platex users
+      #	$latex2e->{latex} = "platex";
+    } elsif ($global->{language} eq "ko") {
+      # for Korean users
+      $latex2e->{latex} = "hlatexp";
+    } else {
+      $latex2e->{latex} = "latex";
+    }
+  }
+  #  Fine tune $latex2e->{latex} for some special cases
+  if ($latex2e->{latex} eq "hlatexp") {
+    # for pdf: how about status?(for hlatex and hlatexp)
+    $latex2e->{latex} = "hlatex";       # run hlatex if hlatexp is used
+  } elsif ($latex2e->{output} eq "pdf" && $latex2e->{latex} eq "latex" ) {
+    $latex2e->{latex} = "pdflatex";
+  }
+
+  # Fine tune $latex2e->{dvips} for defaults and special cases
+  if ($latex2e->{dvips} eq "unknown") {
+    if ($global->{language} eq "ja"){
+      if ( $latex2e->{latex} eq "jlatex" ){
+	$latex2e->{dvips} = "dvi2ps";
+      } else {
+	$latex2e->{dvips} = "dvips";
+      }
+    } else {
+      $latex2e->{dvips} = "dvips";
+    }
+  }
+
+  # Fine tune some other corner cases
+  # We use jlatex for Japanese encoded (euc-jp) characters.
+  # pdf support for Japanese are not yet. use ps for the time being.
+  if ($global->{charset} eq "nippon") {
+    $latex2e->{output} = "ps" if ($latex2e->{output} eq "pdf");
+  }
+
+  # Fine tune japanese with utf-8. Must use uplatex and dvipdfmx
+  if ($global->{charset} eq "utf-8") {
+    if ($global->{language} eq "ja"){
+      print STDERR "  - Force using uplatex and dvipdfmx for japanese+utf-8.\n";
+      ldt_log "  - Force using uplate and dvipdfmx for japanese+utf-8.";
+      $latex2e->{latex}  = "uplatex";
+      $latex2e->{dvips}  = "dvipdfmx";
+      # This sets output to pdf, but since behavior is dvips like we
+      # need to cheat the program using "ps" not to break code
+      # behind. This shoult be fixed in a cleaner way.
+      if ( $latex2e->{output} = "pdf" ) {
+	$latex2e->{output} = "ps";
+      }
+    }
+  }
+
+  ldt_log "  ${msgheader} general info after fine tuning:";
+  ldt_log "    format: $global->{format}, charset: $global->{charset}, lang: $global->{language}";
+  ldt_log "    latex: $latex2e->{latex}, dvips: $latex2e->{dvips}, output: $latex2e->{output}, verbosity: $latex2e->{verbosity}";
+
+  return $latex2e;
+};
+
+# -------------------------------------------------------------------
+sub latex2e_clean_tmplatexdir {
+  # -----------------------------------------------------------------
+  # Clean LaTeX temporary directory
+  # -----------------------------------------------------------------
+  my $tmplatexdir = shift;
+  my $tmplatexnam = shift;
+  my $epsfiles    = shift;
+  my $msgheader   = "fmt_latex2e::latex2e_clean_tmplatexdir";
+
+  die "${msgheader}: tmplatexdir undefined or non existent: \"\"\n"
+    unless ( -d $tmplatexdir );
+  die "${msgheader}: tmplatexnam undefined.\n"
+    unless ( $tmplatexnam );
+
+  if ( $global->{debug} ){
+    print "${msgheader}:\n"
+      . " Temporary LaTeX files are in\n"
+      . "  $tmplatexdir\n"
+      . " Please check dir contents. They are to be removed manually.\n";
+  } else {
+    # ldt_log calls useful on LinucDocTools::Utils::remove_tmpfiles failure
+    my @suffixes = qw(aux bbl blg dlog dvi lof log lot out pdf ps tex toc);
+    for my $suffix (@suffixes) {
+      my $file = "$tmplatexnam.$suffix";
+      if ( -e $file ) {
+	ldt_log "    - Removing file \"$file\"";
+	unlink "$file";
+      }
+    }
+    # Clean epsf files if created
+    for my $epsf ( @{$epsfiles} ) {
+      my $tmpepsf = $tmplatexdir . "/" . $epsf;
+      print $tmpepsf . "\n" if ( $global->{debug} );
+      unlink ("$tmpepsf");
+    }
+    ldt_log "    - Removing dir \"$tmplatexdir\"";
+    rmdir ($tmplatexdir) || return -1;
+  }
+};
+
 # ------------------------------------------------------------------
 $latex2e->{postASP} = sub
 # ------------------------------------------------------------------
@@ -247,6 +377,12 @@ $latex2e->{postASP} = sub
   mkdir ($tmplatexdir, 0700)
     or die "Could not create \"$tmplatexdir\" directory.\n";
 
+  # Some ko charset fine tunings. To be removed once utf-8 is ready.
+  $global->{charset} = "euc-kr" if ($global->{language} eq "ko");
+
+  # Set defaults and fine tune $latex2e->{latex,dvips,output}
+  $latex2e = latex2e_fine_tune_options($latex2e);
+
   # check epsfile is specified in source file
   {
     my $epsf;
@@ -284,11 +420,11 @@ $latex2e->{postASP} = sub
       push @texlines, $_;
       # and check for nameurl
       if ( /\\nameurl/ ){
-	($urlid, $urlnam) = ($_ =~ /\\nameurl{(.*)}{(.*)}/);
+	($urlid, $urlnam) = ($_ =~ /\\nameurl\{(.*)\}\{(.*)\}/);
 	print $urlnum . ": " . $urlid . "\n" if ( $global->{debug} );
 
 	$urldef = latex2e_defnam($urlnum) . "url";
-	s/\\nameurl{.*}{.*}/{\\em $urlnam} {\\tt \\$urldef}/;
+	s/\\nameurl\{.*\}\{.*\}/\\emph{$urlnam} \\texttt{\\$urldef}/;
 	push @urlnames, $_;
 	push @urldefines, "\\urldef{\\$urldef} \\url{$urlid}\n";
 	$urlnum++;
@@ -304,8 +440,6 @@ $latex2e->{postASP} = sub
     my $langlit = ISO2English ($global->{language});
     $langlit = ($langlit eq 'english') ? "" : "$langlit";
     my $hlatexopt = "";
-    $global->{charset} = "nippon" if ($global->{language} eq "ja");
-    $global->{charset} = "euc-kr" if ($global->{language} eq "ko");
 
     # Getting document class prefix
     my $classprefix = "";
@@ -324,9 +458,15 @@ $latex2e->{postASP} = sub
     my $babeloptions = $langlit || "english";
     if ( ($global->{charset} eq "nippon")
 	 ||
-	 ($global->{charset} eq "euc-kr")) {
+	 ($global->{charset} eq "euc-kr")
+	 ||
+	 ($global->{language} eq "ja" && $global->{charset} eq "utf-8")
+      ) {
       $babeloptions = '';
     }
+    ldt_log "  ${msgheader} general info:";
+    ldt_log "    format: $global->{format}, charset: $global->{charset}, lang: $global->{language}";
+    ldt_log "    latex: $latex2e->{latex}, classprefix: $classprefix, babeloptions: $babeloptions";
 
     open ($OUTFILE, "> $tmplatexnam.tex")
       or die "fmt_latex2e::postASP: Could not open \"$tmplatexnam.tex\" for write.\n";
@@ -367,12 +507,12 @@ $latex2e->{postASP} = sub
 	    $_ = $_ . "\\makeindex\n" if ($latex2e->{makeindex});
 	  }
 	  # Set correct DTD name
-	  elsif (/^\\usepackage{\@LINUXDOC_DTD\@-sgml}/) {
+	  elsif (/^\\usepackage\{\@LINUXDOC_DTD\@-sgml\}/) {
 	    my $dtd = $global->{"dtd"};
 	    s/\@LINUXDOC_DTD\@/$dtd/;
 	  }
 	  # Set correct babel options
-	  elsif (/^\\usepackage\[\@BABELOPTIONS\@\]{babel}/) {
+	  elsif (/^\\usepackage\[\@BABELOPTIONS\@\]\{babel\}/) {
 	    if ( $babeloptions ) {
 	      s/\@BABELOPTIONS\@/$babeloptions/;
 	    } else {
@@ -387,6 +527,8 @@ $latex2e->{postASP} = sub
 	  elsif ( /\\usepackage\[\@CHARSET\@\]\{inputenc\}/ ) {
 	    if ( $global->{charset} eq "latin" ) {
 	      s/\@CHARSET\@/latin1/;
+	    } elsif ( $global->{charset} eq "utf-8" ) {
+	      s/\@CHARSET\@/utf8/;
 	    } else {
 	      s/^/%%/;
 	    }
@@ -420,18 +562,17 @@ $latex2e->{postASP} = sub
 
     umask $saved_umask;
     copy ("$tmplatexnam.tex", "$filename.tex");
-    if ( ! $global->{debug} ) {
-      unlink ("$tmplatexnam.tex");
-      rmdir ($tmplatexdir) || return -1;
-    }
-
+    latex2e_clean_tmplatexdir($tmplatexdir,$tmplatexnam,\@epsfiles);
     return 0;
   }
 
   # Run LaTeX in nonstop mode so it won't prompt & hang on errors.
-  # Suppress the output of LaTeX on all but the last pass, after
-  # references have been resolved.  This avoids large numbers of
-  # spurious warnings.
+  # Default behavior is to supress LaTeX output on all passes, unless
+  # error, to avoid large number of spurious warnings.
+  # With --verbosity=1 (-V 1) it will be shown info about last run,
+  # once references have been resolved. With '2' info about all passes
+  # will be shown.
+
   my $current_dir;
   chop ($current_dir = `pwd`);
   print $current_dir . "\n" if ( $global->{debug} );
@@ -446,50 +587,43 @@ $latex2e->{postASP} = sub
   # go to the temporary directory
   chdir ($tmplatexdir);
 
+  # Prepare actual LaTeX run(s)
   my ($latexcommand) = "$latex2e->{latex} '\\nonstopmode\\input{$filename.tex}'";
+  ldt_log "  ${msgheader}:latexcommand\n    $latexcommand";
 
-  # We run pdflatex instead of latex if user selected pdf output
-  if ($latex2e->{output} eq "pdf") {
-    $latexcommand = "pdflatex '\\nonstopmode\\input{$filename.tex}'";
+  # Control verbosity for the different LaTeX runs
+  my $latex1suppress  = ' >/dev/null';
+  my $latex2suppress  = ' >/dev/null';
+  my $latex3suppress  = "";
+  my $bibtexsuppress  = "";
+
+  if ( $latex2e->{verbosity} > 1  ) {
+    $latex1suppress  = "";
+    $latex2suppress  = "";
+  } elsif ( $latex2e->{verbosity} == 0 ) {
+    $latex3suppress = ' >/dev/null';
+    $bibtexsuppress =  ' >/dev/null';
+  } elsif ( $latex2e->{quick} ) {
+    $latex1suppress  = "";
   }
 
-  # run hlatex if hlatexp is used
-  # for pdf: how about status?(for hlatex and hlatexp)
-  if ($latex2e->{latex} eq "hlatexp") {
-    #$latex2e->{output} = "ps" if ($latex2e->{output} eq "pdf");
-    $latexcommand = "hlatex '\\nonstopmode\\input{$filename.tex}'";
+  unless ( system ($latexcommand . $latex1suppress) == 0 ){
+    ldt_cat("$filename.log") unless ( $latex1suppress eq "" );
+    die "$msgheader: LaTeX first run problem. Aborting ...\n";
   }
 
-  # We use jlatex for Japanese encoded (euc-jp) characters.
-  # pdf support for Japanese are not yet. use ps for the time being.
-  if ($global->{charset} eq "nippon") {
-    $latex2e->{output} = "ps" if ($latex2e->{output} eq "pdf");
-    $latexcommand = "$latex2e->{latex} '\\nonstopmode\\input{$filename.tex}'";
-  }
-  my ($suppress) = $latex2e->{quick} ? "" : ' >/dev/null';
-
-  system ($latexcommand . $suppress) == 0
-    or die "$msgheader: LaTeX first run problem. Aborting ...\n";
-
   if ( $latex2e->{bibtex} ) {
-    system ( "bibtex $filename" ) == 0
+    system ( "bibtex $filename" . $bibtexsuppress ) == 0
       or print STDERR "$msgheader: Problems when running BibTeX. Ignoring ...\n";
   }
 
   unless ( $latex2e->{quick} ){
-    system ($latexcommand . ' >/dev/null') == 0
+    system ($latexcommand . $latex2suppress) == 0
       or die "$msgheader: LaTeX second run problem. Aborting ...\n";
-    system ( $latexcommand ) == 0
+    system ( $latexcommand . $latex3suppress) == 0
       or die "$msgheader: LaTeX third run problem. Aborting ...\n";
   }
 
-  if ( ! $global->{debug} ) {
-    my @suffixes = qw(log blg aux toc lof lot dlog bbl out);
-    for my $suf (@suffixes) {
-      unlink "$tmplatexnam.$suf";
-    }
-  }
-
   # go back to the working directory
   chdir ($current_dir);
 
@@ -503,18 +637,7 @@ $latex2e->{postASP} = sub
     #        }
     umask $saved_umask;
     copy ("$tmplatexnam.dvi", "$filename.dvi");
-    if ( $global->{debug} ) {
-      print "Temporary files are in $tmplatexdir\n";
-      print "Please check there and remove them manually.\n";
-    } else {
-      unlink ("$tmplatexnam.tex", "$tmplatexnam.dvi");
-      for my $epsf ( @epsfiles ) {
-	$tmpepsf = $tmplatexdir . "/" . $epsf;
-	print $tmpepsf . "\n" if ( $global->{debug} );
-	unlink ("$tmpepsf");
-      }
-      rmdir ($tmplatexdir) || return -1;
-    }
+    latex2e_clean_tmplatexdir($tmplatexdir,$tmplatexnam,\@epsfiles);
     return 0;
   }
 
@@ -528,31 +651,25 @@ $latex2e->{postASP} = sub
     #        }
     umask $saved_umask;
     copy ("$tmplatexnam.pdf", "$filename.pdf");
-    if ( $global->{debug} ) {
-      print "Temporary files are in $tmplatexdir\n";
-      print "Please check there and remove them manually.\n";
-    } else {
-      unlink ("$tmplatexnam.tex", "$tmplatexnam.pdf");
-      for my $epsf ( @epsfiles ) {
-	$tmpepsf = $tmplatexdir . "/" . $epsf;
-	print $tmpepsf . "\n" if ( $global->{debug} );
-	unlink ("$tmpepsf");
-      }
-      rmdir ($tmplatexdir) || return -1;
-    }
+    latex2e_clean_tmplatexdir($tmplatexdir,$tmplatexnam,\@epsfiles);
     return 0;
   }
 
   # convert dvi into ps using dvips command
   chdir ($tmplatexdir);
-  if ($latex2e->{dvips} eq "dvi2ps") {
-    `dvi2ps -q -o $global->{papersize} -c $tmplatexnam.ps $filename.dvi`;
+  my $dvips_command_line;
+  if ($latex2e->{dvips} eq "dvipdfmx") {
+    # dvipdfmx really produces pdf
+    $dvips_command_line = "dvipdfmx -q -p $global->{papersize} -o $tmplatexnam.pdf $filename.dvi";
+  } elsif ($latex2e->{dvips} eq "dvi2ps") {
+    $dvips_command_line = "dvi2ps -q -o $global->{papersize} -c $tmplatexnam.ps $filename.dvi";
   } elsif ($latex2e->{dvips} eq "jdvi2kps") {
-    `jdvi2kps -q -pa $global->{papersize} -o $tmplatexnam.ps $filename.dvi`;
+    $dvips_command_line = "jdvi2kps -q -pa $global->{papersize} -o $tmplatexnam.ps $filename.dvi";
   } else {
-    `dvips -R -q -t $global->{papersize} -o $tmplatexnam.ps $filename.dvi`;
+     $dvips_command_line = "dvips -R -q -t $global->{papersize} -o $tmplatexnam.ps $filename.dvi";
   }
-
+  ldt_log "  ${msgheader}::dvips_command_line\n    $dvips_command_line";
+  `$dvips_command_line`;
   chdir ($current_dir);
 
   # comment out, because this backup action is not documented yet.
@@ -562,22 +679,15 @@ $latex2e->{postASP} = sub
   #      rename ("$filename.ps", "$filename.ps.back");
   #    }
   umask $saved_umask;
-  copy ("$tmplatexnam.ps", "$filename.ps");
-  unlink ("$tmplatexnam.ps");
-  if ( $global->{debug} ) {
-    print "Temporary files are in $tmplatexdir\n";
-    print "Please check there and remove them manually.\n";
+  if ($latex2e->{dvips} eq "dvipdfmx"){
+    copy ("$tmplatexnam.pdf", "$filename.pdf");
+    unlink ("$tmplatexnam.pdf");
   } else {
-    unlink ("$tmplatexnam.tex", "$tmplatexnam.dvi", "$tmplatexnam.ps");
-    for my $epsf ( @epsfiles ) {
-      $tmpepsf = $tmplatexdir . "/" . $epsf;
-      print $tmpepsf . "\n" if ( $global->{debug} );
-      unlink ("$tmpepsf");
-    }
-    rmdir ($tmplatexdir) || return -1;
+    copy ("$tmplatexnam.ps", "$filename.ps");
+    unlink ("$tmplatexnam.ps");
   }
+  latex2e_clean_tmplatexdir($tmplatexdir,$tmplatexnam,\@epsfiles);
   return 0;
-
 };
 
 1;
