@@ -2,18 +2,18 @@
 #
 #  LinuxDocTools.pm
 #
-#  LinuxDoc-Tools driver core. This contains all the basic functionality
-#  we need to control all other components.
+#  LinuxDoc-Tools driver core. This contains all the basic
+#  functionality we need to control all other components.
 #
 #  Copyright  1996, Cees de Groot.
 #  Copyright  2000, Taketoshi Sano
-#  Copyright  2006-2018, Agustin Martin
-# --------------------------------------------------------------------------------
-
+#  Copyright  2006-2020, Agustin Martin
+# -------------------------------------------------------------------
 package LinuxDocTools;
 
 require 5.006;
 use strict;
+use LinuxDocTools::Data::Latin1ToSgml qw{ldt_latin1tosgml};
 
 =head1 NAME
 
@@ -44,7 +44,14 @@ use File::Copy;
 use File::Temp qw(tempdir);
 use File::Basename qw(fileparse);
 use LinuxDocTools::Lang;
-use LinuxDocTools::Utils qw(usage cleanup trap_signals remove_tmpfiles create_temp);
+use LinuxDocTools::Utils qw(
+  cleanup
+  create_temp
+  ldt_log
+  remove_tmpfiles
+  trap_signals
+  usage
+  );
 use LinuxDocTools::Vars;
 
 sub BEGIN
@@ -58,22 +65,22 @@ sub BEGIN
   use lib "$main::DataDir/site";
 }
 
-# -----------------------------------------------------------------------------------
+# -------------------------------------------------------------------
 sub ldt_searchfile {
-# -----------------------------------------------------------------------------------
-# Look for a readable file in the locations. Return first math.
-# -----------------------------------------------------------------------------------
+  # -----------------------------------------------------------------
+  # Look for a readable file in the locations. Return first math.
+  # -----------------------------------------------------------------
   my $files = shift;
   foreach my $file  ( @$files ){
     return $file if -r $file;
   }
 }
 
-# -----------------------------------------------------------------------------------
+# -------------------------------------------------------------------
 sub ldt_getdtd_v1 {
-# -----------------------------------------------------------------------------------
-# Get the dtd
-# -----------------------------------------------------------------------------------
+  # -----------------------------------------------------------------
+  # Get the dtd
+  # -----------------------------------------------------------------
   my $file         = shift;
   my $error_header = "LinuxdocTools::ldt_getdtd_v1";
   my $dtd;
@@ -126,11 +133,11 @@ sub ldt_getdtd_v1 {
   return $dtd;
 }
 
-# -----------------------------------------------------------------------------------
+# -------------------------------------------------------------------
 sub ldt_getdtd_v2 {
-# -----------------------------------------------------------------------------------
-# Second way of getting dtd, fron nsgmls output.
-# -----------------------------------------------------------------------------------
+  # -----------------------------------------------------------------
+  # Second way of getting dtd, fron nsgmls output.
+  # -----------------------------------------------------------------
   my $preaspout    = shift;
   my $error_header = "LinuxdocTools::ldt_getdtd_v2";
   my $dtd2;
@@ -145,86 +152,7 @@ sub ldt_getdtd_v2 {
   return $dtd2;
 }
 
-# -----------------------------------------------------------------------------------
-sub ldt_latin1tosgml {
-# -----------------------------------------------------------------------------------
-# Convert latin1 chars in input filehandle to sgml entities in the returned string
-# -----------------------------------------------------------------------------------
-  my $FILE     = shift;
-  my $sgmlout;
-
-  while (<$FILE>){
-    # Outline these commands later on - CdG
-    #change latin1 characters to SGML
-    #by Farzad Farid, adapted by Greg Hankins
-    s//\&Agrave;/g;
-    s//\&Aacute;/g;
-    s//\&Acirc;/g;
-    s//\&Atilde;/g;
-    s//\&Auml;/g;
-    s//\&Aring;/g;
-    s//\&AElig;/g;
-    s//\&Ccedil;/g;
-    s//\&Egrave;/g;
-    s//\&Eacute;/g;
-    s//\&Ecirc;/g;
-    s//\&Euml;/g;
-    s//\&Igrave;/g;
-    s//\&Iacute;/g;
-    s//\&Icirc;/g;
-    s//\&Iuml;/g;
-    s//\&Ntilde;/g;
-    s//\&Ograve;/g;
-    s//\&Oacute;/g;
-    s//\&Ocirc;/g;
-    s//\&Otilde;/g;
-    s//\&Ouml;/g;
-    s//\&Oslash;/g;
-    s//\&Ugrave;/g;
-    s//\&Uacute;/g;
-    s//\&Ucirc;/g;
-    s//\&Uuml;/g;
-    s//\&Yacute;/g;
-    s//\&THORN;/g;
-    s//\&szlig;/g;
-    s//\&agrave;/g;
-    s//\&aacute;/g;
-    s//\&acirc;/g;
-    s//\&atilde;/g;
-    s//\&auml;/g;
-    s//\&aring;/g;
-    s//\&aelig;/g;
-    s//\&ccedil;/g;
-    s//\&egrave;/g;
-    s//\&eacute;/g;
-    s//\&ecirc;/g;
-    s//\&euml;/g;
-    s//\&igrave;/g;
-    s//\&iacute;/g;
-    s//\&icirc;/g;
-    s//\&iuml;/g;
-    s//\&mu;/g;
-    s//\&eth;/g;
-    s//\&ntilde;/g;
-    s//\&ograve;/g;
-    s//\&oacute;/g;
-    s//\&ocirc;/g;
-    s//\&otilde;/g;
-    s//\&ouml;/g;
-    s//\&oslash;/g;
-    s//\&ugrave;/g;
-    s//\&uacute;/g;
-    s//\&ucirc;/g;
-    s//\&uuml;/g;
-    s//\&yacute;/g;
-    s//\&thorn;/g;
-    s//\&yuml;/g;
-    $sgmlout .= $_;
-  }
-  return $sgmlout;
-}
-
-# ------------------------------------------------------------------------
+# -------------------------------------------------------------------
 
 =item LinuxDocTools::init
 
@@ -239,9 +167,9 @@ formatting backends and C<require>s them.
 
 =cut
 
-# -----------------------------------------------------------------------------------
+# -------------------------------------------------------------------
 sub init {
-# -----------------------------------------------------------------------------------
+# -------------------------------------------------------------------
   trap_signals;
 
   # Register the ``global'' pseudoformat. Apart from the global settings, we
@@ -264,7 +192,8 @@ sub init {
 			    'values' => [ @LinuxDocTools::Lang::Languages ],
 			    short => "l" },
 			  { option => "charset",   type => "l",
-			    'values' => [ "latin", "ascii", "nippon", "euc-kr" ], short => "c" },
+			    'values' => [ "latin", "ascii", "nippon", "euc-kr" , "utf-8"],
+			    short => "c" },
 			  { option => "style",     type => "s", short => "S" },
 			  { option => "tabsize",   type => "i", short => "t" },
 			  # { option => "verbose",   type => "f", short => "v" },
@@ -283,6 +212,7 @@ sub init {
   $global->{define}    = "";
   $global->{debug}     = 0;
   $global->{include}   = "";
+  $global->{logfile}   = '';
   $global->{pass}      = "";
   $global->{InFiles}   = [];
   $global->{fmtlist}   = "";            # List of loaded fmt files
@@ -313,11 +243,12 @@ sub init {
   #     }
   # }
 
-  # --------------------------------------------------------------------------------
+  # -------------------------------------------------------------------
   $global->{preNSGMLS} = sub {
-    # ------------------------------------------------------------------------------
-    #  Define a fallback preNSGMLS. Used when the format is "global" (from sgmlcheck).
-    # ------------------------------------------------------------------------------
+    # -----------------------------------------------------------------
+    #  Define a fallback preNSGMLS. Used when the format is "global"
+    #  (from sgmlcheck).
+    # -----------------------------------------------------------------
     $global->{NsgmlsOpts}   .= " -s ";
     $global->{NsgmlsPrePipe} = "cat $global->{file}";
   };
@@ -358,13 +289,14 @@ modified and, once I<$global->{format}> is known, I<SGMLDECL> is set.
 
 =cut
 
-# ------------------------------------------------------------------------
+# -------------------------------------------------------------------
 sub process_options {
-# ------------------------------------------------------------------------
-  my $progname = shift;
-  my @tmpargs  = @_;
-  my @args     = ();
-  my $format   = '';
+  # -----------------------------------------------------------------
+  my $progname  = shift;
+  my @tmpargs   = @_;
+  my @args      = ();
+  my $format    = '';
+  my $msgheader = "LinuxDocTools::process_options";
 
   # Try getting the format. We need to do this here so process_options
   # knows which is the format and which format options are allowed
@@ -414,6 +346,19 @@ sub process_options {
   # Normalize language string
   $global->{language} = Any2ISO($global->{language});
 
+  # Fine tune japanese and korean charsets when not utf-8
+  if ($global->{charset} ne "utf-8") {
+    if ($global->{language} eq "ja" ){
+      $global->{charset} = "nippon";
+    } elsif ($global->{language} eq "ko"){
+      if ($global->{format} eq "groff") {
+	$global->{charset} = "latin1";
+      } else {
+	$global->{charset} = "euc-kr";
+      }
+    }
+  }
+
   # Setup the SGML environment.
   my @sgmlcatalogs =
     (# SGML iso-entities catalog location in Debian sgml-data package
@@ -443,7 +388,7 @@ sub process_options {
   return @files;
 }
 
-# ------------------------------------------------------------------------
+# -------------------------------------------------------------------
 
 =item LinuxDocTools::process_file
 
@@ -473,12 +418,13 @@ etcetera. See the code for details.
 
 =cut
 
-# ------------------------------------------------------------------------
+# -------------------------------------------------------------------
 sub process_file {
-# ------------------------------------------------------------------------
+  # ----------------------------------------------------------------
   my $file            = $global->{origfile} = shift (@_);
   my $saved_umask     = umask;
   my $error_header    = "LinuxdocTools::process_file";
+  my $fmtopts         = $Formats{$global->{format}};
 
   print "Processing file $file\n";
   umask 0077;
@@ -493,58 +439,90 @@ sub process_file {
   my $dtd = ldt_getdtd_v1("$global->{file}");
   print STDERR "DTD: " . $dtd . "\n" if $global->{debug};
 
-  # Prepare temporary directory
+  # -- Prepare temporary directory
   my $tmpdir    = $ENV{'TMPDIR'} || '/tmp';
   $tmpdir       = tempdir("linuxdoc-tools.XXXXXXXXXX", DIR => "$tmpdir");
 
-  # Set common base name for temp files and temp file names
+  # -- Set common base name for temp files and temp file names
   my $tmpbase   = $global->{tmpbase} = $tmpdir . '/sgmltmp.' . $filename;
   my $precmdout = "$tmpbase.01.precmdout";
   my $nsgmlsout = "$tmpbase.02.nsgmlsout";   # Was $tmpbase.1
   my $preaspout = "$tmpbase.03.preaspout";   # Was $tmpbase.2
   my $aspout    = "$tmpbase.04.aspout";      # Was $tmpbase.3
 
+  # -- Set $global->{logfile} and initialize logfile.
+  $global->{logfile} = "$tmpbase.$global->{format}.log";
+  open (my $LOGFILE, ">", "$global->{logfile}")
+    or die "$error_header: Could not open \"$global->{logfile}\" logfile for write.\n";
+  print $LOGFILE "--- Opening \"$global->{logfile}\" logfile ---\n";
+  close $LOGFILE;
+
+  # -- Write info about some global options
+  ldt_log "--- Begin: Info about global options";
+  foreach ( sort keys %$global ){
+    next if m/fmtlist|InFiles|OPTIONS/;
+    ldt_log "$_:  $global->{$_}";
+  }
+  ldt_log "--- End: Info about global options";
+  ldt_log "$global->{fmtlist}";
+
+  # -- Write info about some backend options
+  ldt_log "--- Begin: Info about backend options";
+  foreach ( sort keys %$fmtopts ){
+    next if m/fmtlist|InFiles|OPTIONS/;
+    ldt_log "$_:  $fmtopts->{$_}";
+  }
+  ldt_log "--- End: Info about backend options";
+
   # Set up the preprocessing command. Conditionals have to be
   # handled here until they can be moved into the DTD, otherwise
   # a validating SGML parser will choke on them.
 
-  # Check if output option for latex is pdf or not
+  # -- Check if output option for latex is pdf or not
   if ($global->{format} eq "latex2e") {
     if ($Formats{$global->{format}}{output} eq "pdf") {
       $global->{define} .= " pdflatex=yes";
     }
   }
 
-  # Set the actual pre-processing command
+  # -- Set the actual pre-processing command
   my($precmd) = "| $global->{sgmlpre} output=$global->{format} $global->{define}";
+  ldt_log "  ${error_header}::precmd:\n    $precmd";
 
-  # Make sure path of file to be processed is in SGML_SEARCH_PATH
+  # -- Make sure path of file to be processed is in SGML_SEARCH_PATH
   $ENV{"SGML_SEARCH_PATH"} .= ":$filepath";
 
-  # You can hack $NsgmlsOpts here, etcetera.
+  # -- You can hack $NsgmlsOpts here, etcetera.
   $global->{NsgmlsOpts}   .= "-D $main::prefix/share/sgml -D $main::DataDir";
   $global->{NsgmlsOpts}   .= "-i$global->{include}" if ($global->{include});
 
   # If a preNSGMLS function is defined in the fmt file, pipe its output to $FILE,
   # otherwise just open $global->{file} as $IFILE
-  # ----------------------------------------------------------------------------
+  # -----------------------------------------------------------------
+  ldt_log "- PreNsgmls stage started.";
   my $IFILE;
   if ( defined $Formats{$global->{format}}{preNSGMLS} ) {
     $global->{NsgmlsPrePipe} = &{$Formats{$global->{format}}{preNSGMLS}};
+    ldt_log  "  ${error_header}::NsgmlsPrePipe: $global->{NsgmlsPrePipe} |";
     open ($IFILE,"$global->{NsgmlsPrePipe} |")
       || die "$error_header: Could not open pipe from $global->{NsgmlsPrePipe}. Aborting ...\n";
   } else {
+    ldt_log "  ${error_header}: No prepipe. Just opening \"$global->{file}\" for read";
     open ($IFILE,"< $global->{file}")
       || die "$error_header: Could not open $global->{file} for reading. Aborting ...\n";
   }
 
-  # Create a temp file with $precmd output
+  # -- Create a temp file with $precmd output
   my $precmd_command    = "$precmd > $precmdout";
+  ldt_log "  ${error_header}::precmd_command:\n    $precmd_command";
 
   open (my $PRECMDOUT, "$precmd_command")
     or die "$error_header: Could not open pipe to $precmdout. Aborting ...\n";
 
-  if ($global->{charset} eq "latin") {
+  # -- Convert latin1 chars to sgml entities for html backend
+  if ( $global->{format} eq "html"
+       && $global->{charset} eq "latin"  ) {
+    ldt_log "  ${error_header}: Converting latin1 chars to sgml entities for html backend";
     print $PRECMDOUT ldt_latin1tosgml($IFILE);
   } else {
     copy($IFILE,$PRECMDOUT);
@@ -552,16 +530,24 @@ sub process_file {
 
   close $IFILE;
   close $PRECMDOUT;
+  ldt_log "- PreNsgmls stage finished.";
 
-  # Process with nsgmls.
-  my $nsgmls_command = "$main::progs->{NSGMLS} $global->{NsgmlsOpts} $ENV{SGMLDECL} $precmdout > $nsgmlsout";
+  ldt_log "- Nsgmls stage started.";
+
+  # -- Pass apropriate envvars to nsgmls to better deal with utf-8
+  my $NSGMLS_envvars = ($global->{charset} eq "utf-8")
+    ? "SP_CHARSET_FIXED=yes SP_ENCODING=utf-8" : "";
+
+  # -- Process with nsgmls.
+  my $nsgmls_command = "$NSGMLS_envvars $main::progs->{NSGMLS} $global->{NsgmlsOpts} $ENV{SGMLDECL} $precmdout > $nsgmlsout";
+  ldt_log "  ${error_header}::nsgmls_command:\n    $nsgmls_command";
   system($nsgmls_command) == 0
-    or die "Error: \"$nsgmls_command\" failed with exit status: ",$? >> 8,"\n";
+    or die "${error_header}: Error: \"$nsgmls_command\" failed with exit status: ",$? >> 8,"\n";
 
-  #  Special case: if format is global, we're just checking.
+  #  -- Special case: if format is global, we're just checking.
   cleanup if ( $global->{format} eq "global");
 
-  #  If output file does not exists or is empty, something went wrong.
+  # -- If output file does not exists or is empty, something went wrong.
   if ( ! -e "$nsgmlsout" ) {
     die "$error_header: Can't create file $nsgmlsout. Aborting ...\n";
   } elsif ( -z "$nsgmlsout" ){
@@ -569,9 +555,11 @@ sub process_file {
   }
 
   print "- Nsgmls stage finished.\n" if $global->{debug};
+  ldt_log "- Nsgmls stage finished.";
 
   #  If a preASP stage is defined, let the format handle it.
   #  --------------------------------------------------------
+  ldt_log "- PreASP stage started.";
   open (my $PREASP_IN, "< $nsgmlsout")
     or die "$error_header: Could not open $nsgmlsout for reading. Aborting ...\n";
   open (my $PREASP_OUT, "> $preaspout")
@@ -592,12 +580,15 @@ sub process_file {
     unless -e "$preaspout";
 
   print "- PreASP stage finished.\n" if ( $global->{debug} );
+  ldt_log "- PreASP stage finished.";
 
   # Run sgmlsasp, with an optional style if specified.
   # -----------------------------------------------------------
+  ldt_log "- ASP stage started.";
   my $dtd2 = ldt_getdtd_v2($preaspout)
     or die "$error_header: Could not read dtd from $preaspout. Aborting ...\n";
 
+  ldt_log "  $error_header: dtd_v1: $dtd, dtd_v2: $dtd2, both must match, dtd_v2 prevails";
   unless ( $dtd eq $dtd2 ){
     print STDERR "Warning: Two different values for dtd, dtd1: $dtd, dtd2: $dtd2\n";
     $dtd = $dtd2;
@@ -623,17 +614,6 @@ sub process_file {
 
   $mapping = "$style $mapping" if $style;
 
-  $global->{charset} = "nippon" if ($global->{language} eq "ja");
-
-  # We don't have Korean groff so charset should be latin1.
-  if ($global->{language} eq "ko") {
-    if ($global->{format} eq "groff") {
-      $global->{charset} = "latin1";
-    } else {
-      $global->{charset} = "euc-kr";
-    }
-  }
-
   if ($global->{format} eq "groff"){
     if ($dtd eq "linuxdoctr") {
       $mapping = "$main::DataDir/mappings/$global->{format}/tr-mapping";
@@ -642,6 +622,7 @@ sub process_file {
 
   my $sgmlsasp_command = "$main::progs->{SGMLSASP} $mapping < $preaspout |
       expand -t $global->{tabsize} > $aspout";
+  ldt_log "  ${error_header}::sgmlsasp_command:\n    $sgmlsasp_command";
   system ($sgmlsasp_command) == 0
     or die "$error_header: Error running $sgmlsasp_command. Aborting ...\n";
 
@@ -649,9 +630,11 @@ sub process_file {
     unless -e "$aspout";
 
   print "- ASP stage finished.\n" if ( $global->{debug} );
+  ldt_log "- ASP stage finished.";
 
   #  If a postASP stage is defined, let the format handle it.
   # ----------------------------------------------------------------
+  ldt_log "- postASP stage started.";
   umask $saved_umask;
 
   open (my $INPOSTASP, "< $aspout" )
@@ -665,8 +648,12 @@ sub process_file {
   close $INPOSTASP;
 
   print "- postASP stage finished.\n" if ( $global->{debug} );
+  ldt_log "- postASP stage finished.";
 
-  #  All done, remove the temporaries.
+  # -- Reset $global->{logfile} for next file
+  $global->{logfile} = '';
+
+  # -- All done, remove the temporaries.
   remove_tmpfiles($tmpbase) unless ( $global->{debug} );
 }
 
